from . import BaseCallbacker


class AuxillaryCallbacker(BaseCallbacker):
    def update(self, lr, epoch, **kwargs):
        if self.val_mode:
            return
        self.update_values({'aux/lr': lr, 'aux/epoch': epoch})



from .base_callbacker import BaseCallbacker


class LossCallbacker(BaseCallbacker):
    def update(self, outputs, **kwargs):
        if not self.val_mode:
            self.update_values({'loss/training': outputs['loss']})
        else:
            self.update_values({'loss/validation': outputs['loss']})

        if 'losses' in outputs.keys():
            losses = outputs['losses']
            self.update_values({
                f"{task}/{'validation' if self.val_mode else 'training'}/{head}/loss":
                    losses[task][head] for task in losses.keys() for head in losses[task]
            })



import logging
from tqdm import tqdm
import torch

from .base_trainer import BaseTrainer
from trainify.configs import SystemConfig


class Trainer(BaseTrainer):
    def train_step(self, batch, accumulate, accumulation_size):
        if SystemConfig.val_log_freq and self.absolute_step % SystemConfig.val_log_freq == 0:
            if self.absolute_step != 0 or SystemConfig.validate_at_start:
                self._write_val_logs()

        outputs = self.model(**batch, accumulate=accumulate)

        loss = outputs['loss']
        if torch.isnan(loss).any():
            raise ValueError('Loss is nan')
        outputs['loss'] = loss.detach()
        if accumulate:
            loss /= accumulate
        elif accumulation_size:
            loss /= accumulation_size

        if SystemConfig.use_amp:
            import apex
            with apex.amp.scale_loss(loss, self.optimizer) as scaled_loss:
                scaled_loss.backward()
        else:
            loss.backward()
        del loss

        if not accumulate:
            self.optimizer.step()
            if self.scheduler is not None:
                self.scheduler.step()
            self.optimizer.zero_grad()

        if SystemConfig.train_metrics_freq and self.absolute_step % SystemConfig.train_metrics_freq == 0:
            self.callbackers.update(training_mode=True, outputs=outputs, epoch=self.epoch, lr=self.lr)

        if (SystemConfig.checkpoint_step_freq and (self.absolute_step % SystemConfig.checkpoint_step_freq == 0)
                and self.absolute_step != 0):
            self.checkpointer.save(self.model, self.optimizer, self.epoch, self.absolute_step)

        if SystemConfig.train_log_freq and self.absolute_step % SystemConfig.train_log_freq == 0:
            self._write_train_logs()

        self.absolute_step += 1

    def train_core(self):
        if self.enable_logging:
            logging.info('Training started')
        self.model.train()
        if self.start_step is not None:
            assert self.start_step >= 0, \
                'Wrong restoration step, checkpoint was probably trained with different configs'
        else:
            self.start_step = -1
        accumulation_size = None
        for self.epoch in range(self.epoch, SystemConfig.epochs + 1):
            for batch, accumulate in tqdm(
                    self.train_dataloader, desc=f'Training epoch {self.epoch}', disable=not self.enable_logging):
                if self.absolute_step <= self.start_step:
                    self.absolute_step += 1
                    continue
                self.train_step(batch, accumulate, accumulation_size)
                accumulation_size = accumulate

            if hasattr(self.model, 'reset'):
                self.model.reset()

            if SystemConfig.checkpoint_epoch_freq and self.epoch % SystemConfig.checkpoint_epoch_freq == 0:
                self.checkpointer.save(self.model, self.optimizer, self.epoch)

    def train(self):
        try:
            self.train_core()
        except Exception as e:
            if not isinstance(e, KeyboardInterrupt) and self.absolute_step > 2:
                self.checkpointer.save(self.model, self.optimizer, self.epoch, self.absolute_step)
            raise



from . import BasePreparator


class Remover(BasePreparator):
    def __init__(self, columns):
        self.drop_columns = [columns] if isinstance(columns, str) else columns

    def prepare(self, dataset, meta):
        drop_columns = [c for c in self.drop_columns if c in dataset.columns]
        dataset = dataset.drop(columns=drop_columns)
        return dataset, meta



from . import BasePreparator


class Dropper(BasePreparator):
    def __init__(self, n: int = None, frac: float = None, replace: bool = False):
        self.frac = frac
        self.n = n
        assert self.frac is None or self.n is None
        assert self.frac is not None or self.n is not None
        self.replace = replace

    def prepare(self, dataset, meta):
        return dataset.sample(frac=self.frac, n=self.n, replace=self.replace).reset_index(drop=True), meta


class Shuffler(Dropper):
    def __init__(self):
        super().__init__(frac=1, replace=False)



from abc import ABC, abstractmethod
from typing import Optional, Tuple, List, Union
import pandas as pd


class BasePreparator(ABC):
    @abstractmethod
    def prepare(self, dataset: Optional[pd.DataFrame], meta: Optional[dict]) -> Tuple[pd.DataFrame, dict]:
        pass


Preparators = Union[BasePreparator, List['Preparators'], Tuple['Preparators', ...]]



from typing import Optional, Set, DefaultDict
from abc import ABC, abstractmethod
from collections import defaultdict

from ..utils.averagers import Averager, MeanAverager


class BaseCallbacker(ABC):
    def __init__(self, averager_cls: Optional[type(Averager)] = None):
        self.val_mode = False
        self.callbacks: DefaultDict[str, Averager] = defaultdict(MeanAverager if averager_cls is None else averager_cls)
        self.active_callbacks: Set[str] = set()

    def train(self):
        if self.val_mode:
            self.val_mode = False
            self.reset()

    def eval(self):
        if not self.val_mode:
            self.val_mode = True
            self.reset()

    def get_values(self) -> dict:
        return {key: self.callbacks[key].get() for key in self.active_callbacks}

    @abstractmethod
    def update(self, **kwargs):
        pass

    def update_values(self, update_dict: dict):
        for key, value in update_dict.items():
            self.callbacks[key].upd(value)
            self.active_callbacks.add(key)

    def write(self, tb_writer, step):
        for name, value in self.get_values().items():
            if tb_writer is not None:
                tb_writer.add_scalar(name, value, step)

    def reset(self):
        self.active_callbacks = set()
        for callback in self.callbacks.values():
            callback.reset()



from .base_callbacker import BaseCallbacker
from .loss_callbacker import LossCallbacker
from .accuracy_callbacker import AccuracyCallbacker
from .aux_callbacker import AuxillaryCallbacker
from .model_callbacker import ModelCallbacker



from typing import Tuple, List, Dict, Optional, Union
from abc import ABC, abstractmethod
from collections import defaultdict
import os
import random
import pandas as pd
import torch

from trainify.configs import SystemConfig


class BaseSampler(ABC):
    @abstractmethod
    def sample(self, batch_size) -> Dict[str, torch.Tensor]:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def get_input_names(self) -> List[str]:
        pass

    @abstractmethod
    def set_rank(self, rank: int):
        pass


class TensorDataset:
    def __init__(self, tensors_dict: Dict[str, torch.Tensor]):
        sizes = [len(value) for value in tensors_dict.values()]
        assert all(size == sizes[0] for size in sizes)
        self.length = sizes[0]
        self.tensors_dict = tensors_dict

    def to(self, device: Union[str, torch.device]):
        self.tensors_dict = {key: value.to(device) for key, value in self.tensors_dict.items()}

    @property
    def inputs_list(self) -> List[str]:
        return list(self.tensors_dict.keys())

    @staticmethod
    def convert_type(x: torch.Tensor) -> torch.Tensor:
        type_map = {
            int_type: torch.long for int_type in (torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64)}
        return x.to(type_map[x.dtype]) if x.dtype in type_map.keys() else x

    def __getitem__(self, index) -> Dict[str, torch.Tensor]:
        return {
            key: self.convert_type(value[index]).to(SystemConfig.device) for key, value in self.tensors_dict.items()}

    def __len__(self) -> int:
        return self.length

    def share_memory_(self):
        for value in self.tensors_dict.values():
            value.share_memory_()


class OMSampler(BaseSampler):
    def __init__(
            self, tensors_dict: Dict[str, torch.Tensor], shuffle: bool = True, device=None, seed: Optional[int] = None):
        self.dataset = TensorDataset(tensors_dict)
        if device is not None:
            self.to(device)
        self.length = len(self.dataset)
        self.shuffle = shuffle
        self.seed = seed
        self.indices = self._create_indices()
        self.position = 0

    def to(self, device: Union[str, torch.device]):
        self.dataset.to(device)

    def reset(self):
        if self.seed is not None:
            self.seed += 1
        self.position = 0
        self.indices = self._create_indices()

    @property
    def get_input_names(self) -> List[str]:
        return self.dataset.inputs_list

    def __len__(self) -> int:
        return self.length

    def _create_indices(self) -> List[int]:
        indices = list(range(self.length))
        if not self.shuffle:
            return indices

        if self.seed is not None:
            random.seed(self.seed)

        random.shuffle(indices)
        return indices

    def sample(self, batch_size: int) -> Dict[str, torch.Tensor]:
        old_position = self.position
        self.position += batch_size
        if self.position > self.length:
            self.reset()
            raise StopIteration
        return self.dataset[self.indices[old_position: self.position]]

    def share_memory_(self):
        self.dataset.share_memory_()


class StorageSampler(BaseSampler):
    sep = SystemConfig.storage_sep
    file_sep = SystemConfig.storage_file_sep
    key = SystemConfig.hdf_key
    ext = SystemConfig.hdf_ext

    def __init__(self, filename: str):
        self.length = None
        self.table_iterator = None
        self.device = None
        self.scalar_inputs = self.vector_inputs = None
        self.dtypes = None
        self.iter = None
        self.filename = filename
        self.rank = None

    def init(self):
        self.filename = f'{self.filename}{self.ext}' if self.rank is None \
            else f'{self.filename}{self.file_sep}{self.rank}{self.ext}'
        assert os.path.exists(self.filename), f'{self.filename} does not exist'
        store = pd.HDFStore(self.filename)
        self.length = store.get_storer(self.key).attrs.len // SystemConfig.batch_size
        self.table_iterator = store.select(self.key, chunksize=SystemConfig.batch_size)
        self.device = SystemConfig.device
        sample = store.select(self.key, start=0, stop=1)
        assert isinstance(sample, pd.DataFrame)
        self.scalar_inputs, self.vector_inputs = self.read_input_names(columns=list(sample.columns), sep=self.sep)
        raw_dtypes = sample.dtypes
        self.dtypes = {}
        for key, dtype in raw_dtypes.items():
            dtype = self.convert_dtype(dtype.name)
            assert isinstance(key, str)
            key = key.split(self.sep)[0]
            if key in self.dtypes:
                assert self.dtypes[key] == dtype
            else:
                self.dtypes[key] = dtype

        assert set(self.get_input_names()) == set(self.dtypes.keys()), 'Internal error: dtypes not read properly'

        self.iter = iter(self.table_iterator)

    @staticmethod
    def convert_dtype(dtype: str):
        if 'bool' in dtype:
            return torch.bool
        elif 'int' in dtype:
            return torch.long
        elif 'float' in dtype:
            return torch.float
        else:
            raise ValueError(f"Dtype {dtype} can't be recognized")

    @staticmethod
    def read_input_names(columns: List[str], sep: str) -> Tuple[List[str], Dict[str, int]]:
        scalar_inputs = []
        vector_inputs = defaultdict(int)
        for key in columns:
            if sep in key:
                core, ind = key.split(sep)
                vector_inputs[core] = max(vector_inputs[core], int(ind))
            else:
                scalar_inputs.append(key)
        vector_inputs = {key: max_ind + 1 for key, max_ind in vector_inputs.items()}
        return scalar_inputs, vector_inputs

    def get_input_names(self) -> List[str]:
        if self.table_iterator is None:
            self.init()
        return self.scalar_inputs + list(self.vector_inputs.keys())

    def set_rank(self, rank: int):
        self.rank = rank

    def sample(self, batch_size: Optional[int] = None) -> Dict[str, torch.Tensor]:
        if self.table_iterator is None:
            self.init()
        try:
            source = next(self.iter).to_dict(orient='list')
        except StopIteration:
            self.reset()
            raise

        inputs = {key: torch.tensor(source[key], device=self.device, dtype=self.dtypes[key]) for key in self.scalar_inputs}

        for key, max_ind in self.vector_inputs.items():
            inputs[key] = torch.stack(
                [torch.tensor(source[f'{key}{self.sep}{ind}'], device=self.device, dtype=self.dtypes[key])
                 for ind in range(max_ind)], dim=-1)
        return inputs

    def reset(self):
        self.iter = iter(self.table_iterator)

    def __len__(self):
        if self.table_iterator is None:
            self.init()
        return self.length

    def to(self, device: Union[str, torch.device]):
        self.device = device



from . import BasePreparator


class Renamer(BasePreparator):
    def __init__(self, rename_dict: dict):
        self.rename_dict = rename_dict

    def prepare(self, dataset, meta):
        dataset = dataset.rename(columns=self.rename_dict)
        if 'pad_map' in meta.keys():
            for key, new_key in self.rename_dict.items():
                if key in meta['pad_map'].keys():
                    meta['pad_map'][new_key] = meta['pad_map'][key]
                    del meta['pad_map'][key]
        return dataset, meta



from typing import Type
import logging
import os
import pickle as pkl
import torch
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.multiprocessing as mp

from .base_trainer import BaseTrainer

import trainify.configs as configs
from trainify.configs import SystemConfig

exp_environ_key = 'EXP_NAME'


def train_func(rank, trainer):
    logging.info(f'Spawned process with rank {rank}')
    trainer.train_distr(rank)


def ParallelTrainerWrapper(trainer_cls: Type[BaseTrainer]) -> Type[BaseTrainer]:
    def __init__(self, model, optimizer, scheduler, dataloaders, callbackers, checkpointer, start_epoch, start_step):
        trainer_cls.__init__(
            self, model, None, scheduler, dataloaders, callbackers, checkpointer, start_epoch, start_step)

        self.num_gpus = SystemConfig.num_gpus
        self.gpus_per_model = SystemConfig.gpus_per_model

        # optimizer pickled to avoid error caught in mp.spawn after restoring optimizer.state_dict
        # in torch/multiprocessing/reductions.py reduce_storage, line 333
        # RuntimeError: unable to open shared memory object in read-write mode
        # probably because there isn't enough shared memory
        self.optimizer_cls = type(optimizer)
        self.scheduler_cls = type(scheduler) if scheduler is not None else None
        os.makedirs('tmp', exist_ok=True)
        torch.save(optimizer.state_dict(), 'tmp/optimizer.pkl')

    class DDPX(DDP):
        def _ddp_init_helper(self):
            for attr in ('save_pretrained', 'reset'):
                if hasattr(self.module, attr):
                    setattr(self, attr, getattr(self.module, attr))
            self.device_ids = None
            super()._ddp_init_helper()

        def forward(self, *inputs, accumulate=False, **kwargs):
            return self.module(*inputs, **kwargs) if accumulate or not self.module.training \
                else super().forward(*inputs, **kwargs)

    if SystemConfig.distributed_type == 'apex':
        from apex.parallel import DistributedDataParallel as ApexDDP

        class ApexDDPX(ApexDDP):
            def __init__(self, *inputs, **kwargs):
                super().__init__(*inputs, **kwargs)
                for attr in ('save_pretrained', 'reset'):
                    if hasattr(self.module, attr):
                        setattr(self, attr, getattr(self.module, attr))

            def forward(self, *inputs, accumulate=False, **kwargs):
                outputs = super().forward(*inputs, **kwargs)
                if not self.module.training:
                    loss = outputs['loss'].detach()
                    dist.reduce(tensor=loss, dst=0, op=dist.ReduceOp.SUM)
                    outputs['loss'] = loss
                return outputs

    def train_distr(self, rank):
        logging.info(f'Initializing process with rank {rank}')

        local_rank = rank
        if SystemConfig.node_rank is not None:
            SystemConfig.num_nodes = int(SystemConfig.num_nodes)
            SystemConfig.node_rank = int(SystemConfig.node_rank)
            rank += SystemConfig.num_gpus * SystemConfig.node_rank

        dist.init_process_group(
            backend='nccl', init_method=SystemConfig.init_method, rank=rank, world_size=SystemConfig.world_size)

        self.devices = [SystemConfig.parallel_devices[self.gpus_per_model * local_rank + ind]
                        for ind in range(self.gpus_per_model)]
        self.device = self.devices[0]
        SystemConfig.devices = self.devices
        SystemConfig.device = self.device

        if hasattr(self.model, 'to_devices_'):
            self.model.to_devices_(self.devices)
        elif SystemConfig.gpus_per_model == 1:
            self.model = self.model.to(self.devices[0])
        else:
            raise ValueError('multi-gpu models should have to_devices_(devices) method defined')

        self.optimizer = self.optimizer_cls(
            self.model.optimizer_parameters() if hasattr(self.model, 'optimizer_parameters')
            else self.model.parameters(), lr=SystemConfig.lr, **SystemConfig.optimizer_kwargs)

        self.scheduler = self.scheduler_cls(self.optimizer) if self.scheduler_cls is not None else None

        torch.cuda.set_device(self.devices[0])

        if SystemConfig.use_amp:
            import apex
            self.model = apex.parallel.convert_syncbn_model(self.model)
            self.model, self.optimizer = apex.amp.initialize(
                self.model, self.optimizer,
                opt_level=SystemConfig.amp_opt_level, max_loss_scale=SystemConfig.amp_max_loss_scale, min_loss_scale=1)
            with open('tmp/optimizer.pkl', 'rb') as file:
                self.optimizer.load_state_dict(pkl.load(file))
        else:
            self.optimizer.load_state_dict(torch.load('tmp/optimizer.pkl', map_location='cpu'))

        if SystemConfig.distributed_type == 'apex':
            self.model = ApexDDPX(self.model, gradient_predivide_factor=SystemConfig.num_gpus * SystemConfig.batch_size)
        else:
            self.model = DDPX(self.model, find_unused_parameters=SystemConfig.find_unused_parameters)

        self.model.train()

        if self.train_dataloader is not None:
            self.train_dataloader.set_rank(rank)
        if self.val_dataloaders is not None:
            for val_dataloader in self.val_dataloaders.values():
                val_dataloader.set_rank(rank)
        if rank != 0:
            self.callbackers.active = False
            self.checkpointer.active = False
            self.enable_logging = False
        else:
            os.remove('tmp/optimizer.pkl')

        trainer_cls.train(self)

    def train(self):
        os.environ[exp_environ_key] = SystemConfig.exp
        os.environ['MASTER_ADDR'] = SystemConfig.master_addr
        os.environ['MASTER_PORT'] = str(SystemConfig.master_port)
        logging.info('Spawning train processes')
        mp.spawn(train_func, args=(self,), nprocs=self.num_gpus // self.gpus_per_model, join=True)

    global ParallelTrainer
    ParallelTrainer = type(
        'ParallelTrainer',
        (trainer_cls,),
        {'__init__': __init__, 'train': train, 'train_distr': train_distr})

    return ParallelTrainer


exp = os.environ.get(exp_environ_key)
if exp is not None and SystemConfig.exp is None:
    configs.init(exp)



from . import BaseCallbacker


class ModelCallbacker(BaseCallbacker):
    def update(self, outputs, **kwargs):
        if 'callbacks' in outputs.keys():
            self.update_values({f'model_{"validation" if self.val_mode else "training"}/{callback}': value
                                for callback, value in outputs['callbacks'].items()})



from typing import Tuple, Dict, Optional
import os
from shutil import copyfile
import logging
import pandas as pd
import tqdm
import torch

from trainify.preparators import BasePreparator, Preparators, Cacher
from .samplers import StorageSampler
from .dataloaders import SingleTaskLoader
from .checkpointer import Checkpointer

from trainify.configs import SystemConfig, Components


def prepare_dataset(
        preparators: Preparators, dataset: Optional[pd.DataFrame] = None,
        meta: Optional[dict] = None, is_train: Optional[bool] = None) -> Tuple[pd.DataFrame, dict]:
    if meta is None:
        meta = {}
    if is_train is not None:
        meta['is_train'] = is_train
    if not preparators:
        return dataset, meta
    if isinstance(preparators, BasePreparator):
        return preparators.prepare(dataset, meta)
    elif isinstance(preparators, list):
        for ind, sub_prep in reversed(list(enumerate(preparators))):
            if isinstance(sub_prep, Cacher) and sub_prep.ready():
                try:
                    dataset, meta = sub_prep.prepare(dataset, meta)
                    preparators = preparators[ind + 1:]
                    break
                except Exception as e:
                    print(e)
                    sub_prep.rewrite = True
        for sub_prep in preparators:
            dataset, meta = prepare_dataset(sub_prep, dataset, meta)
    elif isinstance(preparators, tuple):
        datasets = []
        metas = []
        for sub_prep in preparators:
            new_dataset, new_meta = prepare_dataset(sub_prep, dataset, meta)
            datasets.append(new_dataset)
            metas.append(new_meta)
        dataset = pd.concat(datasets)
        for new_meta in metas:
            meta.update(new_meta)
    else:
        raise ValueError(
            f'type should be instance or nested list / dict of BasePreparator, received {type(preparators)}')

    if not isinstance(dataset, pd.DataFrame):
        raise ValueError(f'preparators should have returned pandas DataFrame, got {type(dataset)} instead')
    return dataset, meta


def split_dataset(inputs_df: pd.DataFrame) -> Tuple[pd.DataFrame, pd.DataFrame]:
    bsz = SystemConfig.world_batch_size
    val_amount = int(round(len(inputs_df) * SystemConfig.val_size))
    val_amount = min(len(inputs_df) - bsz, max(val_amount, bsz))
    val_amount = val_amount // bsz * bsz
    train_amount = (len(inputs_df) - val_amount) // bsz * bsz
    return inputs_df[:train_amount].reset_index(drop=True), inputs_df[-val_amount:].reset_index(drop=True)


def convert_df_fields(df: pd.DataFrame) -> pd.DataFrame:
    from operator import itemgetter
    for column in list(df.columns):
        logging.info(f'Processing {column} column')
        sample_value = df.loc[0, column]
        if not hasattr(sample_value, '__len__'):
            continue
        length = len(sample_value)
        sep = '||'
        for ind in tqdm.trange(length):
            df[f'{column}{sep}{ind}'] = df[column].apply(itemgetter(ind))

        df = df.drop(columns=[column])
    return df


def init_logging(logs_path: str):
    os.makedirs(logs_path)
    copyfile(os.path.join('configs', f'{SystemConfig.exp_path}.py'), os.path.join(logs_path, 'configs.py'))
    logs_file_path = os.path.join(logs_path, SystemConfig.logs_file)
    logging.basicConfig(
        level=logging.INFO, format=SystemConfig.log_format, datefmt=SystemConfig.log_time_format,
        handlers=[logging.FileHandler(logs_file_path), logging.StreamHandler()])


def create_checkpointer(logs_path: str) -> Checkpointer:
    checkpoints_path = os.path.join(logs_path, SystemConfig.checkpoints_folder)
    os.makedirs(checkpoints_path)

    checkpointer = Checkpointer(checkpoints_path)
    logging.info(f'Checkpointer initialized at {checkpoints_path}')
    return checkpointer


def hdf_save(dataset: pd.DataFrame, filename: str):
    key = SystemConfig.hdf_key
    dataset.to_hdf(filename, key=key, mode='w', format='table')
    with pd.HDFStore(filename, mode='a') as store:
        store.get_storer(key).attrs.len = len(dataset)


def save_dataset(dataset: pd.DataFrame, filename: str):
    dataset = convert_df_fields(dataset)
    if SystemConfig.shuffle_samples:
        dataset = dataset.sample(frac=1).reset_index(drop=True)
    extension = SystemConfig.hdf_ext
    bsz = SystemConfig.batch_size
    world_bsz = SystemConfig.world_batch_size
    if SystemConfig.distributed:
        world_size = SystemConfig.world_size
        for rank in range(world_size):
            worker_indices = []
            for ind in range(0, (len(dataset) // world_bsz) * world_bsz, world_bsz):
                worker_indices += list(range(ind + rank * bsz, ind + (rank + 1) * bsz))
            worker_filename = f'{filename}{SystemConfig.storage_file_sep}{rank}{extension}'
            worker_dataset = dataset.iloc[worker_indices]
            hdf_save(worker_dataset, worker_filename)
    else:
        filename = filename + extension
        hdf_save(dataset, filename)


def prepare_data() -> Tuple[SingleTaskLoader, Optional[Dict[str, SingleTaskLoader]]]:
    preparators = Components.preparators
    samplers = {}

    for task in SystemConfig.tasks:
        logging.info(f'Preparing {task} data')
        path = os.path.join(SystemConfig.data_path, SystemConfig.storage_path, task)
        train_fn = os.path.join(path, 'train' if not SystemConfig.skip_validation else 'train_all')
        val_fn = os.path.join(path, 'val')
        if SystemConfig.rewrite_storage or not os.path.exists(path):
            os.makedirs(path, exist_ok=True)
            if SystemConfig.use_ready_valid_split:
                train_dataset, _ = prepare_dataset(preparators[task], is_train=True)
                val_dataset, _ = (prepare_dataset(preparators[task], is_train=False) if not SystemConfig.skip_validation
                                  else (None, None))
                for dataset, part in ((train_dataset, 'train'), (val_dataset, 'validation')):
                    if dataset is not None and len(dataset) < SystemConfig.world_batch_size:
                        raise ValueError(f'Processed {part} dataset is shorter than 1 batch, '
                                         f'decrease batch size or increase data amount')
            else:
                processed_dataset, _ = prepare_dataset(preparators[task])
                if len(processed_dataset) < 2 * SystemConfig.world_batch_size:
                    raise ValueError('Processed dataset is shorter than 2 batches, '
                                     'decrease batch size or increase data amount')
                train_dataset, val_dataset = (split_dataset(processed_dataset) if not SystemConfig.skip_validation
                                              else (processed_dataset, None))

            save_dataset(train_dataset, train_fn)
            if val_dataset is not None:
                save_dataset(val_dataset, val_fn)
        samplers[task] = StorageSampler(train_fn), StorageSampler(val_fn) if not SystemConfig.skip_validation else None
        logging.info(f'Prepared {task} data')

    # Temporarily works only for single task, refactor in future
    assert len(SystemConfig.tasks) == 1
    task = SystemConfig.tasks[0]
    train_dataloader = Components.dataloader(samplers[task][0])
    val_dataloaders = {task: Components.dataloader(samplers[task][1])} if not SystemConfig.skip_validation else None
    return train_dataloader, val_dataloaders


def create_model(meta: Optional[dict] = None) -> torch.nn.Module:
    model_cls = Components.model
    model = model_cls.from_pretrained(SystemConfig.model_path) if hasattr(model_cls, 'from_pretrained') else model_cls()

    if meta and hasattr(model, 'post_init'):
        model.post_init(**meta)

    if not SystemConfig.distributed:
        if hasattr(model, 'to_devices_'):
            model.to_devices_(SystemConfig.devices)
        elif SystemConfig.gpus_per_model == 1:
            model = model.to(SystemConfig.devices[0])
        else:
            raise ValueError('multi-gpu models should have to_devices_(devices) method defined')

    logging.info('Model created')
    return model


def create_optimizer(model: torch.nn.Module) -> Components.optimizer:
    optimizer = Components.optimizer(
        model.optimizer_parameters() if hasattr(model, 'optimizer_parameters')
        else model.parameters(), lr=SystemConfig.lr, **SystemConfig.optimizer_kwargs)
    if SystemConfig.use_amp and not SystemConfig.distributed:
        import apex
        model, optimizer = apex.amp.initialize(
            model, optimizer, opt_level=SystemConfig.amp_opt_level,
            max_loss_scale=SystemConfig.amp_max_loss_scale)
    logging.info('Optimizer created')
    return optimizer



from typing import Dict, List
import logging
import operator
from collections import defaultdict
from math import ceil
import torch

from .samplers import BaseSampler

from trainify.configs import SystemConfig


class SingleTaskLoader:
    def __init__(self, sampler: BaseSampler):
        self.sampler = sampler
        self.accumulation_size = SystemConfig.total_size // SystemConfig.batch_size
        self.cur_ind = 0

    def reset(self):
        self.sampler.reset()
        self.cur_ind = 0

    def set_rank(self, rank: int):
        self.sampler.set_rank(rank)

    def __iter__(self):
        return self

    def __len__(self):
        return len(self.sampler)

    def __next__(self):
        batch = self.sampler.sample(SystemConfig.batch_size)
        accumulate = self.accumulation_size if self.cur_ind > 0 else 0
        self.cur_ind = (self.cur_ind + 1) % self.accumulation_size
        return batch, accumulate


class MultiDataLoader:
    def __init__(self, samplers: Dict[str, BaseSampler]):
        self.samplers = samplers

        self.common_inputs = [
            inp for inp in next(iter(self.samplers.values())).get_input_names()
            if all(mult_inp != inp and mult_inp not in inp.split('/') for mult_inp in SystemConfig.multitask_inputs)]

        for inp in self.common_inputs:
            for task, sampler in samplers.items():
                if inp not in sampler.get_input_names():
                    raise ValueError(
                        f'input {inp} not included in Config.multitask_inputs but not present for task {task}')

        self.task_inputs = {task: [inp for inp in sampler.get_input_names() if inp not in self.common_inputs]
                            for task, sampler in self.samplers.items()}

        self.tasks = list(self.samplers.keys())
        self.tasks_amount = len(self.samplers)

        self.loader_lengths = {task: sampler.__len__() for task, sampler in self.samplers.items()}
        self.remaining_lengths = self.loader_lengths.copy()

        self.batch_size = SystemConfig.batch_size
        self.total_size = SystemConfig.total_size
        if self.total_size % self.batch_size:
            raise ValueError('total_size should be either None or proportional to batch_size')

        if SystemConfig.distributed:
            self.rank = None
            self.world_size = SystemConfig.world_size
            self.step = 0

        self.cur_size = self.total_size
        self.active = {task: True for task in self.tasks}

        self.remaining_samples = {}
        self.delayed_activations = []
        self.delayed_deactivations = []
        self.size_changed = False

    @property
    def active_tasks(self) -> List[str]:
        return [task for task in self.tasks if self.active[task]]

    @property
    def accumulate(self) -> bool:
        return self.remaining_samples and sum(self.remaining_samples.values()) > 0

    def activate_task(self, task: str):
        if not self.accumulate:
            self.active[task] = True
        else:
            self.delayed_activations.append(task)

    def deactivate_task(self, task: str):
        if not self.accumulate:
            self.active[task] = False
        else:
            self.delayed_deactivations.append(task)

    def activate_all_tasks(self):
        if not self.accumulate:
            self.active = {task: True for task in self.tasks}
        else:
            self.delayed_activations = self.tasks

    def deactivate_all_tasks(self):
        if not self.accumulate:
            self.active = {task: False for task in self.tasks}
        else:
            self.delayed_deactivations = self.tasks

    def set_rank(self, rank: int):
        self.rank = rank

    def _split_batch_by_tasks(self):
        if not self.accumulate \
                and (sum([self.remaining_lengths[task] for task in self.active_tasks]) < self.total_size
                     or any(self.remaining_lengths[task] < 1 for task in self.active_tasks)):
            self.finish()
        total_length = sum([self.remaining_lengths[task] for task in self.active_tasks])
        task_proportions = {task: self.remaining_lengths[task] / total_length for task in self.active_tasks}
        task_sizes = {task: int(round(self.cur_size * task_proportion))
                      for task, task_proportion in task_proportions.items()}
        while sum(task_sizes.values()) > self.cur_size:
            task_sizes[max(task_sizes.items(), key=operator.itemgetter(1))[0]] -= 1
        while sum(task_sizes.values()) < self.cur_size:
            task_sizes[max(task_sizes.items(), key=operator.itemgetter(1))[0]] += 1
        for task in self.active_tasks:
            if task_sizes[task] < 1:
                self.cur_size = ((ceil(1 / min(task_proportions.values())) - 1) // self.batch_size + 1) \
                                * self.batch_size
                if not self.size_changed:
                    logging.info('Size changed to ', self.cur_size)
                    self.size_changed = True
                if self.cur_size > sum(self.remaining_lengths[task] for task in self.active_tasks):
                    self.finish()
                return self._split_batch_by_tasks()
        return task_sizes

    def _next_batch_by_task(self, task: str, size: int):
        assert self.remaining_lengths[task] >= size
        self.remaining_lengths[task] -= size
        return self.samplers[task].sample(batch_size=size)

    def _subsample_batch(self):
        new_sizes = {}
        samples_left = self.batch_size
        for task in self.active_tasks:
            if self.remaining_samples[task] == 0:
                continue
            if self.remaining_samples[task] > samples_left:
                new_sizes[task] = samples_left
                self.remaining_samples[task] -= samples_left
                break
            else:
                new_sizes[task] = self.remaining_samples[task]
                self.remaining_samples[task] = 0
                samples_left -= new_sizes[task]
        assert sum(new_sizes.values()) == self.batch_size
        return new_sizes

    def _get_task_sizes(self) -> Dict[str, int]:
        if not self.accumulate:
            for task in self.delayed_activations:
                self.activate_task(task)
            self.delayed_activations = []
            for task in self.delayed_deactivations:
                self.deactivate_task(task)
            self.delayed_deactivations = []

        if self.accumulate:
            return self._subsample_batch()
        self.cur_size = self.total_size
        task_sizes = self._split_batch_by_tasks()
        if self.cur_size != self.batch_size:
            self.remaining_samples = task_sizes.copy()
            task_sizes = self._subsample_batch()
        return task_sizes

    def _get_batch(self, task_sizes):
        if not SystemConfig.distributed:
            return {task: self._next_batch_by_task(task, size) for task, size in task_sizes.items()}, task_sizes

        if self.rank is None:
            raise ValueError('Rank not set in dataloader')

        worker_ind = self.rank  # (self.rank + self.step) % self.world_size
        self.step += 1
        worker_range = (
            self.batch_size // self.world_size * worker_ind,
            self.batch_size // self.world_size * (worker_ind + 1))
        total_ind = 0
        batch = {}
        worker_task_sizes = {}
        for task, size in task_sizes.items():
            task_batch = self._next_batch_by_task(task, size)
            if total_ind + size < worker_range[0] or total_ind >= worker_range[1]:
                total_ind += size
                continue
            task_range = range(max(worker_range[0] - total_ind, 0), min(worker_range[1] - total_ind, size))
            if len(task_range) == 0:
                total_ind += size
                continue
            worker_task_sizes[task] = len(task_range)
            batch[task] = {
                key: ({k: v[task_range] for k, v in value.items()} if type(value) is dict else value[task_range])
                for key, value in task_batch.items()}
            total_ind += size
        return batch, worker_task_sizes

    def reset(self):
        for sampler in self.samplers.values():
            sampler.reset()
        self.remaining_lengths = self.loader_lengths.copy()
        self.remaining_samples = None
        self.cur_size = self.total_size

    def finish(self):
        self.reset()
        raise StopIteration

    def __iter__(self):
        return self

    def __len__(self):
        return sum([self.loader_lengths[task] for task in self.active_tasks]) // self.batch_size

    @classmethod
    def get_nested_dict(cls, inp_dict: dict) -> dict:
        nested_dict = defaultdict(dict)
        for key, value in inp_dict.items():
            if '/' in key:
                outer_key, inner_key = key.split(sep='/', maxsplit=1)
                nested_dict[outer_key].update(cls.get_nested_dict({inner_key: value}))
            else:
                nested_dict[key] = value
        return dict(nested_dict)

    def __next__(self):
        if not self.active_tasks:
            raise ValueError('No tasks activated')

        task_sizes = self._get_task_sizes()
        batch, task_sizes = self._get_batch(task_sizes)

        task_ids = {}
        cur_ind = 0
        for task, size in task_sizes.items():
            task_ids[task] = list(range(cur_ind, cur_ind + size))
            cur_ind += size

        new_batch = defaultdict(dict, {
            inp: torch.cat([batch[task][inp]
                            for task in batch.keys()], dim=0) if len(batch) > 1 else next(iter(batch.values()))[inp]
            for inp in self.common_inputs})

        for task, inputs in self.task_inputs.items():
            if task in batch.keys():
                for inp in inputs:
                    new_batch[inp][task] = batch[task][inp]

        batch = self.get_nested_dict(new_batch)

        return task_ids, batch, self.cur_size // self.batch_size if self.accumulate else 0

    def share_memory_(self):
        for sampler in self.samplers.values():
            if hasattr(sampler, 'share_memory_'):
                sampler.share_memory_()


class DatasetCoeffsLoader(MultiDataLoader):
    def __init__(self, samplers: dict):
        super().__init__(samplers)
        self.dataset_coeffs = SystemConfig.dataset_coeffs
        for task, coeff in self.dataset_coeffs.items():
            self.loader_lengths[task] = int(round(self.loader_lengths[task] * coeff))
        self.remaining_lengths = self.loader_lengths.copy()

    def _next_batch_by_task(self, task, size):
        assert self.remaining_lengths[task] >= size
        self.remaining_lengths[task] -= size
        try:
            return self.samplers[task].sample(batch_size=size)
        except StopIteration:
            return self.samplers[task].sample(batch_size=size)

    def reset(self):
        self.remaining_lengths = self.loader_lengths.copy()
        self.remaining_samples = None
        self.cur_size = self.total_size



from typing import Tuple
import logging


class ConfigMetaClass(type):
    def __new__(mcs, name: str, bases: Tuple[type, ...], attrs_dict: dict) -> type:
        if not bases:
            config_class = super().__new__(mcs, name, bases, attrs_dict)
            config_class.__initialized__ = False
            return config_class
        for base in bases:
            assert isinstance(base, ConfigMetaClass), f'{name} inherited from non-config class {base}'
            assert not base.__initialized__, f'{name} trying to override already initialized {base}'
        unused_keys = []
        for key, value in attrs_dict.items():
            if not key.startswith('_'):
                is_unused = True
                for base in bases:
                    if hasattr(base, key):
                        setattr(base, key, value)
                        is_unused = False
                if is_unused:
                    unused_keys.append(key)
        if unused_keys:
            logging.warning(f'Attributes {unused_keys} of {name} were not used, rename to _[name] to suppress warning')
        for base in bases:
            if hasattr(base, 'init'):
                base.init()
        return super().__new__(mcs, name, bases, attrs_dict)



import os

import torch
from torch.optim.optimizer import Optimizer

from trainify.configs import SystemConfig


class Checkpointer:
    def __init__(self, path):
        if not os.path.exists(path):
            raise ValueError(f'Checkpointer root path {path} does not exist')
        self.root = path
        self.active = True

    def epoch_step_path(self, epoch, step=None):
        return os.path.join(self.root, f'epoch_{epoch}{(f"_{step}" if step else "")}')

    def model_checkpoint_fn(self, epoch, step=None):
        return os.path.join(self.epoch_step_path(epoch, step), 'model.pt')

    def optimizer_checkpoint_fn(self, epoch, step=None):
        return os.path.join(self.epoch_step_path(epoch, step), 'optimizer.pt')

    def amp_checkpoint_fn(self, epoch, step=None):
        return os.path.join(self.epoch_step_path(epoch, step), 'amp.pt')

    def save(self, model: torch.nn.Module, optimizer: Optimizer, epoch: int, step: int = None):
        if not self.active:
            return
        os.makedirs(self.epoch_step_path(epoch, step))
        torch.save(
            {k.replace('module.', ''): v for k, v in model.state_dict().items()}, self.model_checkpoint_fn(epoch, step))
        torch.save(optimizer.state_dict(), self.optimizer_checkpoint_fn(epoch, step))
        if SystemConfig.use_amp and not SystemConfig.distributed:
            import apex
            torch.save(apex.amp.state_dict(), self.amp_checkpoint_fn(epoch, step))

        # duplicated save for out-of-the-box use for huggingface transformer models
        if hasattr(model, 'save_pretrained'):
            model_path = os.path.join(self.epoch_step_path(epoch, step), 'model')
            os.makedirs(model_path)
            model.save_pretrained(model_path)

    def load(self, epoch, step=None):
        model_state_dict = torch.load(self.model_checkpoint_fn(epoch, step), map_location='cpu')
        optimizer_state_dict = torch.load(self.optimizer_checkpoint_fn(epoch, step), map_location='cpu')
        if SystemConfig.use_amp and not SystemConfig.distributed:
            import apex
            apex.amp.load_state_dict(torch.load(self.amp_checkpoint_fn(epoch, step), map_location='cpu'))
        return model_state_dict, optimizer_state_dict



from typing import Optional, Tuple
import os
import logging
import pickle as pkl
import pandas as pd

from . import BasePreparator

from trainify.configs import SystemConfig


class Loader(BasePreparator):
    """
    base class for loading datasets. may be used as it is, or extended to support wider variety of formats
    """
    mode_exts = {'hdf': ('h5', 'hdf', 'hdf5'), 'pkl': ('pkl', 'pickle', 'pt')}
    ext_to_mode = {ext: mode for mode, exts in mode_exts.items() for ext in exts}
    modes = tuple(mode_exts.keys())
    key = SystemConfig.hdf_key

    base_path = SystemConfig.dataset_path

    def __init__(self, filename: str, meta_filename: Optional[str] = None, mode: str = None):
        if meta_filename is not None:
            logging.warning('using meta_filename parameter should be avoided, reserved mostly for debugging purposes')
        if mode is None:
            if '.' in filename:
                mode = self.ext_to_mode.get(filename.rsplit('.', maxsplit=1)[-1], None)
            if mode is None:
                mode = 'pkl'
                logging.debug("Loader mode couldn't be recognized by filename, setting to pickle")
        else:
            if '.' in filename:
                ext = filename.rsplit('.', maxsplit=1)[-1]
                if self.ext_to_mode.get(ext, mode) != mode:
                    raise ValueError(f'Loader mode {mode} is conflicting with extension {ext}')
            else:
                filename = f'{filename}.{mode}'
        assert mode in self.modes, f'mode is {mode}, available: {self.modes}'

        filename = os.path.join(self.base_path, filename)
        if SystemConfig.use_ready_valid_split:
            split_fn = filename.rsplit('.', maxsplit=1)
            fn_part, ext = split_fn[0], split_fn[1]
            train_part = 'train' if not SystemConfig.skip_validation else 'train_all'
            self.filenames = {True: f'{fn_part}_{train_part}.{ext}', False: f'{fn_part}_valid.{ext}'}
        else:
            self.filenames = {None: filename}
        self.meta_filename = os.path.join(self.base_path, meta_filename) if meta_filename else None
        self.mode = mode

    def load(self, is_train: Optional[bool]) -> Tuple[pd.DataFrame, dict]:
        filename = self.filenames[is_train]
        assert os.path.exists(filename), f"Loader couldn't find file {filename}"
        if self.mode == 'pkl':
            with open(filename, 'rb') as file:
                inp = pkl.load(file)
            if type(inp) is tuple:
                dataset, meta = inp
            else:
                dataset = inp
                meta = {}
        elif self.mode == 'hdf':
            dataset = pd.read_hdf(filename)
            with pd.HDFStore(filename) as store:
                attrs = store.get_storer(self.key).attrs
                meta = attrs.meta if hasattr(attrs, 'meta') else {}
        else:
            raise AssertionError(f'Internal Loader error: no load mode {self.mode}')
        if self.meta_filename:
            with open(self.meta_filename, 'rb') as file:
                meta = pkl.load(file)
        if isinstance(dataset, list):
            dataset = pd.DataFrame(dataset)
        return dataset, meta

    def prepare(self, dataset=None, meta=None):
        if dataset is not None:
            raise ValueError(
                f'Loader should be at the begging of the preparators pipeline, instead received {type(dataset)}')
        return self.load(is_train=meta.get('is_train'))



from . import BasePreparator


class Filterer(BasePreparator):
    def __init__(self, label, values, remove=False):
        self.label = label
        self.values = values if type(values) is list else [values]
        self.remove = remove

    def prepare(self, dataset, meta):
        filter_lambda = (lambda x: x not in self.values) if self.remove else (lambda x: x in self.values)
        return dataset[dataset.loc[:, self.label].apply(filter_lambda)].reset_index(drop=True), meta



import argparse


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description='Run trainer script, Select an experiment name and provide mu;tinode configs and needed')

    parser.add_argument("--exp", type=str, help="Name of experiment", required=True)

    parser.add_argument('--node-rank', type=int, default=None, help='Rank of the node', required=False)
    parser.add_argument('--num-nodes', type=int, default=None, help='Number of nodes', required=False)
    parser.add_argument('--master-addr', type=str, default=None, help='Master address', required=False)
    parser.add_argument('--master-port', type=int, default=None, help='Master port', required=False)

    return parser


def prepare_configs(args: argparse.Namespace):
    import trainify.configs as configs
    configs.init(args.exp)

    from trainify.configs import SystemConfig
    for key, value in vars(args).items():
        assert hasattr(SystemConfig, key), f'Internal error: parameter {key} not present in SystemConfig'
        if value is not None:
            SystemConfig.key = value


def train():
    parser = create_parser()
    args = parser.parse_args()
    prepare_configs(args)

    from .system.builder import Builder

    trainer = Builder().create_trainer()
    trainer.train()



from trainify.preparators import BasePreparator

from .dropper import Dropper, Shuffler
from .filterer import Filterer
from .adder import Adder
from .remover import Remover
from .padder import Padder
from .renamer import Renamer
from .cacher import Cacher
from .loader import Loader



from abc import ABC, abstractmethod
import operator
from collections import defaultdict
import torch

from trainify.configs import SystemConfig


class Averager(ABC):
    def __init__(self, initial_value=0, reduce=True):
        self.initial_value = initial_value
        self.reduce = reduce

    @abstractmethod
    def upd(self, val):
        pass

    @abstractmethod
    def calculate(self):
        pass

    def get(self):
        value = self.calculate()
        if self.reduce and SystemConfig.distributed and isinstance(value, torch.Tensor):
            import torch.distributed as dist
            dist.reduce(tensor=value, dst=0, op=dist.ReduceOp.SUM)
            value /= SystemConfig.world_size
        return value

    @abstractmethod
    def reset(self):
        pass

    def __str__(self):
        return self.get()


class MeanAverager(Averager):
    def __init__(self, initial_value=0.):
        super().__init__(initial_value)
        self.sum = self.initial_value
        self.cnt = 1 if self.initial_value else 0

    def upd(self, val):
        self.sum += val
        self.cnt += 1

    def calculate(self):
        return self.sum / self.cnt if self.cnt != 0 else 0

    def reset(self):
        self.sum = self.initial_value
        self.cnt = 1 if self.initial_value else 0


class MostOftenAverager(Averager):
    def __init__(self, initial_value=0.):
        super().__init__(initial_value, reduce=False)
        self.value_counts = defaultdict(int)

    def upd(self, val):
        self.value_counts[val.item() if isinstance(val, torch.Tensor) else val] += 1

    def calculate(self):
        return max(
            self.value_counts.items(), key=operator.itemgetter(1))[0] if self.value_counts else self.initial_value

    def reset(self):
        self.value_counts = defaultdict(int)


class ExpAverager(Averager):
    """
    Exponential moving average
    """

    def __init__(self, alpha, initial_value=0):
        super().__init__(initial_value)
        self.sum = self.initial_value
        self.alpha = alpha

    def upd(self, val):
        self.sum = self.alpha * self.sum + (1 - self.alpha) * val

    def calculate(self):
        return self.sum

    def reset(self):
        self.sum = self.initial_value



from . import BasePreparator


class Adder(BasePreparator):
    def __init__(self, label, value):
        self.label = label
        self.value = value

    def prepare(self, dataset, meta):
        dataset[self.label] = self.value
        return dataset, meta



from typing import Optional, List, Dict, Union, Any
import torch

from . import ConfigMetaClass


class SystemConfig(metaclass=ConfigMetaClass):
    # exp stores the value of --exp key argument passed to train.py in module-like representation - [exp_group.]exp
    exp: Optional[str] = None
    # exp_path stores the same value represented as a relative config path - [exp_group/]exp
    exp_path: Optional[str] = None

    # DATASETS

    # subfolder where input datasets are stored
    dataset_path = 'datasets'
    # subfolder for cached preprocessing data
    data_path = 'data'
    # train and validation datasets are separated beforehand
    use_ready_valid_split = False

    # subfolder where processed storages are saved
    storage_path = 'storage'
    # ignore and rewrite processed storages if they already exist
    rewrite_storage = False
    # filename separator for multi-worker datasets
    storage_file_sep = '_'
    # storage file extension
    hdf_ext = '.h5'
    # internal storage value for naming scalars got by splitting vector values elementwise
    storage_sep = '||'
    # base key used for IO operations with HDF files
    hdf_key = 'df'

    # LOGGING & CHECKPOINTS PATHS

    # path for logs to be saved to
    logs_path = 'logs'
    # name of logs subfolder for the experiment - if set to None will be same as experiment name
    logs_name: Optional[str] = None
    # log file name
    logs_file = 'run.log'
    # model path
    model_path: Optional[str] = None
    # checkpoint folder name
    checkpoints_folder = 'checkpoints'
    # logging format
    log_format = '%(levelname)s - %(module)s:%(funcName)s:%(lineno)s - %(asctime)s - %(message)s'
    date_format = '%y-%m-%d'
    time_format = '%H-%M-%S'
    log_time_format = f'{date_format}:{time_format}'

    # LOGGING

    train_metrics_freq = 1
    train_log_freq = 64
    val_log_freq = 1024
    checkpoint_step_freq: Optional[int] = None
    checkpoint_epoch_freq = 1

    pickle_protocol = 3

    # TASKS

    tasks: List[str] = ['main']

    weighted_tasks: List[str] = []

    # TRAINING

    batch_size = 1
    world_batch_size = 1
    total_size: Optional[int] = None
    epochs = 1

    use_amp = False
    amp_opt_level: Optional[str] = None
    amp_max_loss_scale: float = 2 ** 8

    optimizer_kwargs: Dict[str, Any] = {}
    lr = 0.001

    use_gradient_checkpointing = False

    skip_validation = False
    validate_at_start = True
    val_size = 0.025

    # DATA PREPARATION

    shuffle_samples = True

    multitask_inputs: List[str] = []

    # padding lengths for sequential inputs
    pad_len = 128

    dataset_coeffs = {}

    # DEVICES

    devices: Union[List[str], List[torch.device]] = ['cpu']
    device = devices[0]
    parallel_devices: List[torch.device] = None
    gpus_per_model = 1
    distributed = False

    distributed_type = 'ddp'

    # additional config for 'ddp'
    find_unused_parameters = True
    # additional config for 'apex'
    apex_max_loss_scale = 2. ** 8

    # RESTORING FROM CHECKPOINT

    restore = False
    restore_epoch = None
    restore_from = None
    restore_step = None

    # MULTINODE TRAINING

    node_rank: Optional[int] = None
    num_nodes: Optional[int] = None

    num_gpus = 1
    world_size = 1
    init_method = 'env://'

    master_addr = 'localhost'
    master_port = 8000

    @classmethod
    def init(cls):
        if isinstance(cls.devices, str):
            cls.devices = [cls.devices]

        if 'cpu' in cls.devices and cls.use_amp:
            raise ValueError(
                'Apex amp is only compatible with gpu devices, modify either Config.use_amp or Config.devices')

        cls.devices = [torch.device(device) for device in cls.devices]
        cls.device = cls.devices[0]

        if len(cls.tasks) == 1:
            cls.task = cls.tasks[0]

        if cls.total_size is None:
            cls.total_size = cls.batch_size

        if cls.total_size % cls.batch_size:
            raise ValueError('If set, Config.total_size should be a multiple of Config.batch_size')

        if cls.skip_validation:
            cls.val_log_freq = None
            cls.validate_at_start = False

        if cls.train_log_freq % cls.train_metrics_freq:
            raise ValueError('Config.train_log_freq should be a multiple of Config.train_metrics_freq')

        if cls.use_amp and (not hasattr(cls, 'amp_opt_level') or cls.amp_opt_level not in 'O0, O1, O2, O3'):
            raise ValueError('When using apex amp, Config.amp_opt_level should be set to one of "O0", "O1", "O2", "O3"')

        if cls.gpus_per_model > len(cls.devices):
            raise ValueError('Amount of devices set in Config.device is smaller than Config.gpus_per_model')

        cls.distributed = (len(cls.devices) > cls.gpus_per_model)

        if cls.distributed:
            cls.num_gpus = len(cls.devices)
            cls.parallel_devices = cls.devices
            cls.devices = [torch.device('cpu')]
            cls.device = cls.devices[0]
            cls.world_size = cls.num_gpus // cls.gpus_per_model
            assert cls.world_size > 0

            cls.world_batch_size = cls.batch_size * cls.world_size
            cls.total_size = cls.total_size * cls.world_size

            if cls.distributed_type == 'ddp' and not hasattr(cls, 'find_unused_parameters'):
                cls.find_unused_parameters = True

            if cls.distributed_type == 'ddp' and cls.use_amp and cls.amp_opt_level == 'O2':
                raise ValueError("Apex amp opt_level O2 isn't compatible with pytorch.distributed DDP")

            if cls.distributed_type == 'apex' and cls.gpus_per_model > 1:
                raise ValueError('Apex DDP supports only single-device models')

            if cls.node_rank is not None:
                if cls.num_nodes is None:
                    raise ValueError('When passing node_rank, num_nodes should be also set')
                cls.init_method = f'{cls.master_addr}:{cls.master_port}'
                cls.world_size *= cls.num_nodes



from typing import Optional, List
import torch

from . import ConfigMetaClass, SystemConfig

from trainify.system.utils.dataloaders import SingleTaskLoader
from trainify.system.trainers.trainer import Trainer
from trainify.system.callbackers import (BaseCallbacker, LossCallbacker, AccuracyCallbacker,
                                                   AuxillaryCallbacker, ModelCallbacker)

from trainify.preparators import Preparators, Cacher


def recursive_cacher_check(preparators: Preparators, paths: Optional[set] = None):
    if paths is None:
        paths = set()
    if isinstance(preparators, list) or isinstance(preparators, tuple):
        for sub_prep in preparators:
            recursive_cacher_check(sub_prep, paths)
    elif isinstance(preparators, dict):
        for sub_prep in preparators.values():
            recursive_cacher_check(sub_prep, paths)
    elif isinstance(preparators, Cacher):
        for filename in preparators.filenames.values():
            if filename in paths:
                raise ValueError(f'Multiple Cachers use same filename: {filename}')
            paths.add(filename)


class Components(metaclass=ConfigMetaClass):
    model: Optional[torch.nn.Module] = None
    optimizer = torch.optim.Adam
    scheduler = None
    preparators = None
    dataloader = SingleTaskLoader
    trainer: type(Trainer) = Trainer
    callbackers: List[BaseCallbacker] = [
        LossCallbacker(), AccuracyCallbacker(), AuxillaryCallbacker(), ModelCallbacker()]

    @classmethod
    def init(cls):
        if SystemConfig.distributed:
            from trainify.system.trainers import ParallelTrainerWrapper
            cls.trainer = ParallelTrainerWrapper(cls.trainer)

        if len(SystemConfig.tasks) == 1 and not isinstance(cls.preparators, dict):
            cls.preparators = {SystemConfig.tasks[0]: cls.preparators}

        recursive_cacher_check(cls.preparators)



import importlib

from .config_metaclass import ConfigMetaClass
from .system_config import SystemConfig
from .default_components import Components


def init(exp: str):
    exp = exp.replace('/', '.')
    importlib.import_module(f'configs.{exp}')
    SystemConfig.exp = exp
    SystemConfig.exp_path = exp.replace('.', '/')



from abc import abstractmethod
from typing import Tuple, Dict, Optional
from tqdm import tqdm

import torch

from trainify.system.utils.callbacker_module import CallbackerModule
from trainify.configs import SystemConfig

from ..utils.dataloaders import MultiDataLoader
from ..utils.checkpointer import Checkpointer


class BaseTrainer:
    def __init__(
            self, model: torch.nn.Module, optimizer, scheduler,
            dataloaders: Tuple[MultiDataLoader, Dict[str, MultiDataLoader]],
            callbackers: CallbackerModule, checkpointer: Checkpointer, start_epoch: int, start_step: Optional[int]):
        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.train_dataloader, self.val_dataloaders = dataloaders
        self.callbackers = callbackers
        self.checkpointer = checkpointer
        self.start_epoch = start_epoch
        self.start_step = start_step

        self.tasks = SystemConfig.tasks
        self.n_tasks = len(self.tasks)
        self.absolute_step = 0
        self.epoch = self.start_epoch
        self.device = SystemConfig.device
        self.enable_logging = True

    @property
    def lr(self):
        return self.optimizer.param_groups[0]['lr']

    def _write_train_logs(self):
        self.callbackers.write(self.absolute_step)

    def _write_val_logs(self):
        self.model.eval()

        for task, val_dataloader in self.val_dataloaders.items():
            for batch, _ in tqdm(
                    val_dataloader, desc=f'Validation at step {self.absolute_step}', disable=not self.enable_logging):
                with torch.no_grad():
                    outputs = self.model(**batch)
                    self.callbackers.update(training_mode=False, outputs=outputs, epoch=self.epoch, lr=self.lr)
            self.callbackers.write(self.absolute_step)

        self.model.train()

    @abstractmethod
    def train(self):
        pass



from typing import Optional
import logging
import os
import time

from trainify.configs import SystemConfig, Components
from .utils.builder_utils import init_logging, create_checkpointer, prepare_data, create_model, create_optimizer
from .utils.checkpointer import Checkpointer
from trainify.system.utils.callbacker_module import CallbackerModule


class Builder:
    def __init__(self):
        self.base_logs_path = os.path.join(
            SystemConfig.logs_path, SystemConfig.logs_name if SystemConfig.logs_name else SystemConfig.exp_path)

        self.logs_path = os.path.join(
            self.base_logs_path, time.strftime(SystemConfig.date_format), time.strftime(SystemConfig.time_format))

        init_logging(self.logs_path)

        self.checkpointer = create_checkpointer(self.logs_path)
        self.callbackers = CallbackerModule(Components.callbackers, self.logs_path)

        self.dataloaders = prepare_data()

        self.model = create_model()
        self.optimizer = create_optimizer(self.model)
        self.scheduler = (Components.scheduler(self.optimizer, dataloader_length=len(self.dataloaders[0]))
                          if Components.scheduler is not None else None)

        self.start_epoch = 0
        self.start_step: Optional[int] = None

        if SystemConfig.restore:
            self.restore_checkpoint()

    def restore_checkpoint(self):
        self.start_step = SystemConfig.restore_step
        restorator = Checkpointer(os.path.join(SystemConfig.restore_from, SystemConfig.checkpoints_folder))
        restore_epoch = SystemConfig.restore_epoch if SystemConfig.restore_epoch != 'last' \
            else max([int(file.split('.')[0].split('_')[1]) for file in os.listdir(restorator.root)])
        self.start_epoch = restore_epoch if self.start_step is not None else restore_epoch + 1
        model_state_dict, optimizer_state_dict = restorator.load(restore_epoch, self.start_step)
        self.model.load_state_dict(model_state_dict)
        self.optimizer.load_state_dict(optimizer_state_dict)

        logging.info('Model and optimizer restored')

    def create_trainer(self) -> Components.trainer:
        return Components.trainer(
            model=self.model, optimizer=self.optimizer, scheduler=self.scheduler,
            dataloaders=self.dataloaders, callbackers=self.callbackers, checkpointer=self.checkpointer,
            start_epoch=self.start_epoch, start_step=self.start_step)



import torch

from . import BaseCallbacker


class AccuracyCallbacker(BaseCallbacker):
    @staticmethod
    def accuracy(logits: torch.Tensor, labels: torch.Tensor) -> torch.Tensor:
        preds = torch.argmax(logits.view((-1, logits.shape[-1])), dim=-1)
        labels = labels.flatten()
        acc = (preds == labels).sum().float() / len(labels)
        return acc

    def update(self, outputs, **kwargs):
        logits = outputs.get('logits')
        labels = outputs.get('labels')
        assert logits is not None, 'to use AccuracyCallbacker, model should have "logits" key in outputs'
        assert labels is not None, 'to use AccuracyCallbacker, model should have "labels" key in outputs'
        self.update_values({f"{'validation' if self.val_mode else 'training'}/accuracy": self.accuracy(logits, labels)})



from typing import Optional, Tuple
import os
import pickle as pkl
import pandas as pd

from .loader import Loader

from trainify.configs import SystemConfig


class Cacher(Loader):
    base_path = SystemConfig.data_path
    key = SystemConfig.hdf_key

    def __init__(self, filename: str, rewrite: bool = False, mode: Optional[str] = None):
        super().__init__(filename=filename, meta_filename=None, mode=mode)
        self.rewrite = rewrite
        self.pickle_protocol: int = SystemConfig.pickle_protocol

    def ready(self) -> bool:
        return all(os.path.exists(filename) for filename in self.filenames.values()) and not self.rewrite

    @staticmethod
    def create_directory_if_needed(path: str):
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

    def prepare(self, dataset=None, meta=None) -> Tuple[pd.DataFrame, dict]:
        if self.ready():
            return self.load(is_train=meta.get('is_train'))
        assert dataset is not None and meta is not None, \
            "Cacher received None inputs, but couldn't load previously saved data"
        filename = self.filenames[meta.get('is_train')]
        self.create_directory_if_needed(filename)
        if self.mode == 'pkl':
            with open(filename, 'wb') as file:
                pkl.dump((dataset, meta), file, protocol=self.pickle_protocol)
        elif self.mode == 'hdf':
            dataset.to_hdf(filename, key=self.key)
            with pd.HDFStore(filename) as store:
                store.get_storer(self.key).attrs.meta = meta

        return dataset, meta



from typing import List
from torch.utils.tensorboard import SummaryWriter

from trainify.system.callbackers import BaseCallbacker


class CallbackerModule:
    def __init__(self, callbackers: List[BaseCallbacker], logs_path: str):
        self.callbackers = callbackers
        self.logs_path = logs_path
        self.tb_writer = None
        self.active = True

    def update(self, training_mode: bool = True, **kwargs):
        for callbacker in self.callbackers:
            if training_mode:
                callbacker.train()
            else:
                callbacker.eval()
            callbacker.update(**kwargs)

    def write(self, absolute_step: int):
        if self.tb_writer is None and self.active:
            self.tb_writer = SummaryWriter(log_dir=self.logs_path)
        for callbacker in self.callbackers:
            callbacker.write(self.tb_writer, absolute_step)
            callbacker.reset()



from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple, Dict
import tqdm
import pandas as pd

from . import BasePreparator


class MapPreparator(BasePreparator, ABC):
    def __init__(self, desc: Optional[str], delete_old: bool = True):
        self.desc = desc
        self.delete_old = delete_old

    @abstractmethod
    def process_row(self, **kwargs) -> Dict[str, Any]:
        pass

    def _process(self, row: pd.Series, bar: tqdm.tqdm) -> Dict[str, Any]:
        result = self.process_row(**row)
        if self.delete_old:
            for key in row.index:
                row[key] = None
        bar.update()
        return result

    def update_meta(self, meta: Dict[str, any]) -> Dict[str, any]:
        return meta

    def prepare(self, dataset: Optional[pd.DataFrame], meta: Optional[dict]) -> Tuple[pd.DataFrame, dict]:
        with tqdm.tqdm(desc=self.desc, total=len(dataset)) as bar:
            return (dataset.apply(lambda row: self._process(row, bar), axis=1, result_type='expand'),
                    self.update_meta(meta))



import logging

from .base_trainer import BaseTrainer


class SingleValidationTrainer(BaseTrainer):
    def train(self):
        logging.info('Checkpoint evaluation started')
        self._write_val_logs()



import logging
from tqdm import tqdm
import numpy as np

from . import BasePreparator


class Padder(BasePreparator):
    def __init__(self, max_len=None):
        self.max_len = max_len

    def pad_input(self, inp, pad_value, pad_len, bar=None):
        assert len(inp) <= pad_len
        shape = (pad_len,) + (inp.shape[1:] if hasattr(inp, 'shape') else ())
        res = np.ones(shape=shape, dtype=np.long) * pad_value
        res[:len(inp)] = inp
        if bar is not None:
            bar.update()
        return res

    def prepare(self, dataset, meta):
        pad_map = meta['pad_map']
        logging.info('Started padding')
        original_len = len(dataset)
        sequential_input = next(iter(pad_map.keys()))
        if self.max_len is None:
            self.max_len = dataset.loc[:, sequential_input].apply(len).max()
            logging.info(f'Padding length assigned to {self.max_len}')
        else:
            dataset = dataset.loc[dataset.loc[:, sequential_input].apply(
                lambda x: len(x) <= self.max_len)].reset_index(drop=True)
            logging.info(f'Using {len(dataset)} of {original_len} total samples')
        for input_name, pad_value in pad_map.items():
            if np.isnan(pad_value):
                continue
            with tqdm(total=len(dataset), desc=f'Padding {input_name}') as bar:
                dataset.loc[:, input_name] = dataset.loc[:, input_name].apply(
                    lambda x: self.pad_input(x, pad_value, self.max_len, bar))
        return dataset, meta
